# Don't manually change, let poetry-dynamic-versioning-plugin handle it.
__version__ = "0.7.0"

from .config import RegistryConfig
from .exceptions import (
    CannotDeriveNameError,
    CannotRegisterPythonBuiltInError,
    InvalidNameError,
    KeyCollisionError,
    ModuleAliasError,
    RegistryError,
)
from .registry import Registry, RegistryDecorator, RegistryMeta
